@@ -3,7 +3,7 @@
  *
  * @copyright Copyright (C) 2010-2011 Vic Lee.
  * @copyright Copyright (C) 2014-2015 Antenore Gatta, Fabio Castelli, Giovanni Panozzo.
- * @copyright Copyright (C) 2016-2020 Antenore Gatta, Giovanni Panozzo.
+ * @copyright Copyright (C) 2016-2021 Antenore Gatta, Giovanni Panozzo.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -43,6 +43,8 @@
 #include <gtk/gtk.h>
 #include <glib/gi18n.h>
 #include <gio/gio.h>
+#include <glib-object.h>
+#include <gobject/gvaluecollector.h>
 #include <vte/vte.h>
 #include <locale.h>
 #include <langinfo.h>
@@ -55,11 +57,16 @@
 #include "remmina_ssh_plugin.h"
 #include "remmina_masterthread_exec.h"
 
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
+
+
 #define REMMINA_PLUGIN_SSH_FEATURE_TOOL_COPY  1
 #define REMMINA_PLUGIN_SSH_FEATURE_TOOL_PASTE 2
 #define REMMINA_PLUGIN_SSH_FEATURE_TOOL_SELECT_ALL 3
 #define REMMINA_PLUGIN_SSH_FEATURE_TOOL_INCREASE_FONT 4
 #define REMMINA_PLUGIN_SSH_FEATURE_TOOL_DECREASE_FONT 5
+#define REMMINA_PLUGIN_SSH_FEATURE_TOOL_SEARCH 6
 
 #define GET_PLUGIN_DATA(gp) (RemminaPluginSshData *)g_object_get_data(G_OBJECT(gp), "plugin-data");
 
@@ -72,9 +79,9 @@
 
 enum color_schemes { LINUX, TANGO, GRUVBOX, SOLARIZED_DARK, SOLARIZED_LIGHT, XTERM, CUSTOM };
 
-/** 16 color palettes in GdkRGBA format (red, green, blue, alpha).
- * Text displayed in the first 8 colors (0-7) is meek (uses thin strokes).
- * Text displayed in the second 8 colors (8-15) is bold (uses thick strokes).
+/** 16 colour palettes in GdkRGBA format (red, green, blue, alpha).
+ * Text displayed in the first 8 colours (0-7) is meek (uses thin strokes).
+ * Text displayed in the second 8 colours (8-15) is bold (uses thick strokes).
  **/
 const GdkRGBA linux_palette[PALETTE_SIZE] = {
 	{ 0,	    0,	      0,	1 },
@@ -200,6 +207,29 @@ static struct {
 
 
 /** The SSH plugin implementation */
+typedef struct _RemminaSshSearch {
+        GtkWidget* parent;
+
+	GtkBuilder* builder;
+	GtkWidget* window;
+        GtkWidget* search_entry;
+        GtkWidget* search_prev_button;
+        GtkWidget* search_next_button;
+        GtkWidget* close_button;
+        GtkToggleButton* match_case_checkbutton;
+        GtkToggleButton* entire_word_checkbutton;
+        GtkToggleButton* regex_checkbutton;
+        GtkToggleButton* wrap_around_checkbutton;
+        GtkWidget* reveal_button;
+        GtkWidget* revealer;
+
+        //VteTerminal *terminal;
+        gboolean regex_caseless;
+        gboolean has_regex;
+        gchar* regex_pattern;
+
+} RemminaSshSearch;
+
 typedef struct _RemminaPluginSshData {
 	RemminaSSHShell *	shell;
 	GFile *			vte_session_file;
@@ -208,9 +238,13 @@ typedef struct _RemminaPluginSshData {
 	const GdkRGBA *		palette;
 
 	pthread_t		thread;
+
+	RemminaSshSearch *  search_widget;
 } RemminaPluginSshData;
 
 
+#define GET_OBJECT(object_name) gtk_builder_get_object(search_widget->builder, object_name)
+
 static RemminaPluginService *remmina_plugin_service = NULL;
 
 static gboolean
@@ -233,6 +267,7 @@ remmina_plugin_ssh_main_thread(gpointer data)
 	RemminaSSH *ssh;
 	RemminaSSHShell *shell = NULL;
 	gboolean cont = FALSE;
+	gboolean partial = FALSE;
 	gchar *hostport;
 	gint ret;
 
@@ -243,18 +278,21 @@ remmina_plugin_ssh_main_thread(gpointer data)
 	remminafile = remmina_plugin_service->protocol_plugin_get_file(gp);
 
 	/* we may need to open a new tunnel */
+	REMMINA_DEBUG ("Tentatively create an SSH tunnel");
 	hostport = remmina_plugin_service->protocol_plugin_start_direct_tunnel(gp, 22, FALSE);
 	if (hostport == NULL) {
 		remmina_plugin_service->protocol_plugin_signal_connection_closed(gp);
 		return NULL;
 	}
+	REMMINA_DEBUG ("protocol_plugin_start_direct_tunnel returned hostport: %s", hostport);
 
 	ssh = g_object_get_data(G_OBJECT(gp), "user-data");
 	if (ssh) {
-		/* Create SSH shell connection based on existing SSH session */
-
+		REMMINA_DEBUG ("Creating SSH shell based on existing SSH session");
 		shell = remmina_ssh_shell_new_from_ssh(ssh);
+		REMMINA_DEBUG ("Calling remmina_public_get_server_port");
 		remmina_plugin_service->get_server_port(hostport, 22, &ssh->tunnel_entrance_host, &ssh->tunnel_entrance_port);
+		REMMINA_DEBUG ("tunnel_entrance_host: %s, tunnel_entrance_port: %d", ssh->tunnel_entrance_host, ssh->tunnel_entrance_port);
 
 		if (remmina_ssh_init_session(REMMINA_SSH(shell)) &&
 		    remmina_ssh_auth(REMMINA_SSH(shell), NULL, gp, remminafile) == REMMINA_SSH_AUTH_SUCCESS &&
@@ -264,35 +302,70 @@ remmina_plugin_ssh_main_thread(gpointer data)
 	} else {
 
 		/* New SSH Shell connection */
+		REMMINA_DEBUG ("Creating SSH shell based on a new SSH session");
 		shell = remmina_ssh_shell_new_from_file(remminafile);
 		ssh = REMMINA_SSH(shell);
+		REMMINA_DEBUG ("Calling remmina_public_get_server_port");
 		remmina_plugin_service->get_server_port(hostport, 22, &ssh->tunnel_entrance_host, &ssh->tunnel_entrance_port);
+		REMMINA_DEBUG ("tunnel_entrance_host: %s, tunnel_entrance_port: %d", ssh->tunnel_entrance_host, ssh->tunnel_entrance_port);
 
 		while (1) {
-			if (!remmina_ssh_init_session(ssh)) {
-				REMMINA_DEBUG("init session error: %s", ssh->error);
-				remmina_plugin_service->protocol_plugin_set_error(gp, "%s", ssh->error);
-				break;
+			if (!partial) {
+				if (!remmina_ssh_init_session(ssh)) {
+					REMMINA_DEBUG("init session error: %s", ssh->error);
+					remmina_plugin_service->protocol_plugin_set_error(gp, "%s", ssh->error);
+					// exit the loop here: OK
+					break;
+				}
 			}
 
 			ret = remmina_ssh_auth_gui(ssh, gp, remminafile);
-			if (ret != REMMINA_SSH_AUTH_SUCCESS) {
-				if (ret != REMMINA_SSH_AUTH_USERCANCEL)
+			switch (ret) {
+				case REMMINA_SSH_AUTH_SUCCESS:
+					REMMINA_DEBUG ("Authentication success");
+					if (!remmina_ssh_shell_open(shell, (RemminaSSHExitFunc)
+								remmina_plugin_service->protocol_plugin_signal_connection_closed, gp)) {
+						remmina_plugin_service->protocol_plugin_set_error(gp, "%s", ssh->error);
+						break;
+					}
+					break;
+				case REMMINA_SSH_AUTH_PARTIAL:
+					REMMINA_DEBUG ("Continue with the next auth method");
+					partial = TRUE;
+					// Continue the loop: OK
+					continue;
+					break;
+				case REMMINA_SSH_AUTH_RECONNECT:
+					REMMINA_DEBUG ("Reconnecting...");
+					if (ssh->session) {
+						ssh_disconnect(ssh->session);
+						ssh_free(ssh->session);
+						ssh->session = NULL;
+					}
+					g_free(ssh->callback);
+					// Continue the loop: OK
+					continue;
+					break;
+				case REMMINA_SSH_AUTH_USERCANCEL:
+					REMMINA_DEBUG ("Interrupted by the user");
+					// TODO: exit the loop here: OK
+					goto BREAK;
+					break;
+				default:
+					REMMINA_DEBUG ("Error during the authentication: %s", ssh->error);
 					remmina_plugin_service->protocol_plugin_set_error(gp, "%s", ssh->error);
-				break;
+					// TODO: exit the loop here: OK
+					goto BREAK;
 			}
 
-			if (!remmina_ssh_shell_open(shell, (RemminaSSHExitFunc)
-						    remmina_plugin_service->protocol_plugin_signal_connection_closed, gp)) {
-				remmina_plugin_service->protocol_plugin_set_error(gp, "%s", ssh->error);
-				break;
-			}
 
 			cont = TRUE;
 			break;
 		}
 	}
 
+BREAK:REMMINA_DEBUG("Authentication terminted with exit status %d", ret);
+
 	g_free(hostport);
 
 	if (!cont) {
@@ -333,12 +406,6 @@ void remmina_plugin_ssh_vte_terminal_set_encoding_and_pty(VteTerminal *terminal,
 
 	setlocale(LC_ALL, "");
 	if (codeset && codeset[0] != '\0') {
-#if VTE_CHECK_VERSION(0, 38, 0)
-		vte_terminal_set_encoding(terminal, codeset, NULL);
-#else
-		vte_terminal_set_emulation(terminal, "xterm");
-		vte_terminal_set_encoding(terminal, codeset);
-#endif
 	}
 
 	vte_terminal_set_backspace_binding(terminal, VTE_ERASE_ASCII_DELETE);
@@ -400,7 +467,13 @@ remmina_plugin_ssh_set_vte_pref(RemminaProtocolWidget *gp)
 				      pango_font_description_from_string(remmina_pref.vte_font));
 #endif
 	}
+
+#if VTE_CHECK_VERSION (0, 51, 3)
+	REMMINA_DEBUG ("Using vte_terminal_set_bold_is_bright instead of vte_terminal_set_allow_bold as deprecated");
+	vte_terminal_set_bold_is_bright(VTE_TERMINAL(gpdata->vte), remmina_pref.vte_allow_bold_text);
+#else
 	vte_terminal_set_allow_bold(VTE_TERMINAL(gpdata->vte), remmina_pref.vte_allow_bold_text);
+#endif
 	if (remmina_pref.vte_lines > 0)
 		vte_terminal_set_scrollback_lines(VTE_TERMINAL(gpdata->vte), remmina_pref.vte_lines);
 }
@@ -458,8 +531,7 @@ remmina_plugin_ssh_vte_save_session(GtkMenuItem *menuitem, RemminaProtocolWidget
 
 	if (err != NULL) {
 		// TRANSLATORS: %s is a placeholder for an error message
-		widget = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK,
-						_("Error: %s"), err->message);
+		widget = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, _("Error: %s"), err->message);
 		g_signal_connect(G_OBJECT(widget), "response", G_CALLBACK(gtk_widget_destroy), NULL);
 		gtk_widget_show(widget);
 		return;
@@ -494,6 +566,206 @@ static void remmina_ssh_keystroke(RemminaProtocolWidget *gp, const guint keystro
 	return;
 }
 
+
+/* regex */
+
+static void jit_regex(VteRegex* regex, char const* pattern)
+{
+	TRACE_CALL(__func__);
+        GError *error;
+        if (!vte_regex_jit(regex, PCRE2_JIT_COMPLETE, &error) ||
+            !vte_regex_jit(regex, PCRE2_JIT_PARTIAL_SOFT, &error)) {
+                if (!g_error_matches(error, VTE_REGEX_ERROR, -45 /* PCRE2_ERROR_JIT_BADOPTION: JIT not supported */))
+                        REMMINA_DEBUG("JITing regex “%s” failed: %s\n", pattern, error->message);
+        }
+}
+
+static VteRegex* compile_regex_for_search(char const* pattern, gboolean caseless, GError** error)
+{
+	TRACE_CALL(__func__);
+        uint32_t flags = PCRE2_UTF | PCRE2_NO_UTF_CHECK | PCRE2_MULTILINE;
+        if (caseless)
+                flags |= PCRE2_CASELESS;
+
+        VteRegex *regex = vte_regex_new_for_search(pattern, strlen(pattern), flags, error);
+        if (regex != NULL)
+                jit_regex(regex, pattern);
+
+        return regex;
+}
+
+static void
+remmina_search_widget_update_sensitivity(RemminaSshSearch* search_widget)
+{
+	TRACE_CALL(__func__);
+        gboolean can_search = search_widget->has_regex;
+
+        gtk_widget_set_sensitive(search_widget->search_next_button, can_search);
+        gtk_widget_set_sensitive(search_widget->search_prev_button, can_search);
+}
+
+static void
+remmina_search_widget_update_regex(RemminaPluginSshData* gpdata)
+{
+	TRACE_CALL(__func__);
+	GError *error = NULL;
+
+	RemminaSshSearch *search_widget = gpdata->search_widget;
+        char const* search_text = gtk_entry_get_text(GTK_ENTRY(search_widget->search_entry));
+        gboolean caseless = gtk_toggle_button_get_active(search_widget->match_case_checkbutton) == FALSE;
+
+        char* pattern;
+        if (gtk_toggle_button_get_active(search_widget->regex_checkbutton))
+                pattern = g_strdup(search_text);
+        else
+                pattern = g_regex_escape_string(search_text, -1);
+
+        if (gtk_toggle_button_get_active(search_widget->regex_checkbutton)) {
+                char* tmp = g_strdup_printf("\\b%s\\b", pattern);
+                g_free(pattern);
+                pattern = tmp;
+        }
+
+        if (caseless == search_widget->regex_caseless &&
+            g_strcmp0(pattern, search_widget->regex_pattern) == 0)
+                return;
+
+        search_widget->regex_caseless = caseless;
+        g_free(search_widget->regex_pattern);
+        search_widget->regex_pattern = NULL;
+
+        if (search_text[0] != '\0') {
+		REMMINA_DEBUG("Search text is: %s", search_text);
+                VteRegex *regex = compile_regex_for_search(pattern, caseless, &error);
+                vte_terminal_search_set_regex(VTE_TERMINAL(gpdata->vte), regex, 0);
+                if (regex != NULL)
+                        vte_regex_unref(regex);
+
+                if (!error) {
+                        search_widget->has_regex = TRUE;
+                        search_widget->regex_pattern = pattern; /* adopt */
+                        pattern = NULL; /* adopted */
+                        gtk_widget_set_tooltip_text(search_widget->search_entry, NULL);
+                } else {
+                        search_widget->has_regex = FALSE;
+			REMMINA_DEBUG ("Regex not set, cannot search");
+                        gtk_widget_set_tooltip_text(search_widget->search_entry, error->message);
+                }
+        }
+
+        g_free(pattern);
+        g_free(error);
+
+        remmina_search_widget_update_sensitivity(search_widget);
+}
+
+static void
+remmina_search_widget_wrap_around_toggled(GtkToggleButton* button, RemminaPluginSshData* gpdata)
+{
+	TRACE_CALL(__func__);
+
+        vte_terminal_search_set_wrap_around(VTE_TERMINAL(gpdata->vte), gtk_toggle_button_get_active(button));
+}
+
+static void
+remmina_search_widget_search_forward(RemminaPluginSshData* gpdata)
+{
+	TRACE_CALL(__func__);
+
+	RemminaSshSearch *search_sidget = gpdata->search_widget;
+        if (!search_sidget->has_regex)
+                return;
+        vte_terminal_search_find_next(VTE_TERMINAL(gpdata->vte));
+}
+
+static void
+remmina_search_widget_search_backward(RemminaPluginSshData* gpdata)
+{
+	TRACE_CALL(__func__);
+
+	RemminaSshSearch *search_sidget = gpdata->search_widget;
+        if (!search_sidget->has_regex)
+                return;
+        vte_terminal_search_find_previous(VTE_TERMINAL(gpdata->vte));
+}
+
+GtkWidget * remmina_plugin_pop_search_new (GtkWidget *relative_to, RemminaProtocolWidget *gp)
+{
+	TRACE_CALL(__func__);
+	RemminaPluginSshData *gpdata = GET_PLUGIN_DATA(gp);
+
+	gpdata->search_widget = g_new0(RemminaSshSearch, 1);
+	RemminaSshSearch *search_widget = gpdata->search_widget;
+
+        search_widget->regex_caseless = FALSE;
+        search_widget->has_regex = FALSE;
+        search_widget->regex_pattern = NULL;
+
+	search_widget->builder = remmina_public_gtk_builder_new_from_resource("/org/remmina/Remmina/src/../data/ui/remmina_search.glade");
+	search_widget->window = GTK_WIDGET(GET_OBJECT("RemminaSearchWidget"));
+        search_widget->search_entry = GTK_WIDGET(GET_OBJECT("search_entry"));
+        search_widget->search_prev_button = GTK_WIDGET(GET_OBJECT("search_prev_button"));
+        search_widget->search_next_button = GTK_WIDGET(GET_OBJECT("search_next_button"));
+        search_widget->close_button = GTK_WIDGET(GET_OBJECT("close_button"));
+        search_widget->match_case_checkbutton = GTK_TOGGLE_BUTTON(GET_OBJECT("match_case_checkbutton"));
+        search_widget->entire_word_checkbutton = GTK_TOGGLE_BUTTON(GET_OBJECT("entire_word_checkbutton"));
+        search_widget->regex_checkbutton = GTK_TOGGLE_BUTTON(GET_OBJECT("regex_checkbutton"));
+        search_widget->wrap_around_checkbutton = GTK_TOGGLE_BUTTON(GET_OBJECT("wrap_around_checkbutton"));
+        search_widget->reveal_button = GTK_WIDGET(GET_OBJECT("reveal_button"));
+        search_widget->revealer = GTK_WIDGET(GET_OBJECT("revealer"));
+
+	gtk_widget_set_can_default (search_widget->search_next_button, TRUE);
+	gtk_widget_grab_default (search_widget->search_next_button);
+
+	gtk_entry_set_activates_default (GTK_ENTRY(search_widget->search_entry), TRUE);
+
+	/* Connect signals */
+	gtk_builder_connect_signals(search_widget->builder, NULL);
+
+        g_signal_connect_swapped(search_widget->close_button, "clicked", G_CALLBACK(gtk_widget_destroy), GTK_WIDGET(search_widget->window));
+
+        g_object_bind_property(search_widget->reveal_button, "active",
+                               search_widget->revealer, "reveal-child",
+                               G_BINDING_DEFAULT | G_BINDING_SYNC_CREATE);
+
+        g_signal_connect_swapped(search_widget->search_entry, "next-match", G_CALLBACK(remmina_search_widget_search_forward), gpdata);
+        g_signal_connect_swapped(search_widget->search_entry, "previous-match", G_CALLBACK(remmina_search_widget_search_backward), gpdata);
+        g_signal_connect_swapped(search_widget->search_entry, "search-changed", G_CALLBACK(remmina_search_widget_update_regex), gpdata);
+
+        g_signal_connect_swapped(search_widget->search_next_button,"clicked", G_CALLBACK(remmina_search_widget_search_forward), gpdata);
+        g_signal_connect_swapped(search_widget->search_prev_button,"clicked", G_CALLBACK(remmina_search_widget_search_backward), gpdata);
+
+        g_signal_connect_swapped(search_widget->match_case_checkbutton,"toggled", G_CALLBACK(remmina_search_widget_update_regex), gpdata);
+        g_signal_connect_swapped(search_widget->entire_word_checkbutton,"toggled", G_CALLBACK(remmina_search_widget_update_regex), gpdata);
+        g_signal_connect_swapped(search_widget->regex_checkbutton,"toggled", G_CALLBACK(remmina_search_widget_update_regex), gpdata);
+        g_signal_connect_swapped(search_widget->match_case_checkbutton, "toggled", G_CALLBACK(remmina_search_widget_update_regex), gpdata);
+
+        g_signal_connect(search_widget->wrap_around_checkbutton, "toggled", G_CALLBACK(remmina_search_widget_wrap_around_toggled), gpdata);
+
+        remmina_search_widget_update_sensitivity(search_widget);
+	return search_widget->window;
+
+}
+
+void remmina_plugin_pop_search (GtkMenuItem *menuitem, RemminaProtocolWidget *gp)
+{
+	TRACE_CALL(__func__);
+	RemminaPluginSshData *gpdata = GET_PLUGIN_DATA(gp);
+
+	GtkWindow *parent = NULL;
+
+	REMMINA_DEBUG ("Before popover");
+	GtkWidget *window = remmina_plugin_pop_search_new (gpdata->vte, gp);
+	GtkWidget *toplevel = gtk_widget_get_toplevel (gpdata->vte);
+	if (GTK_IS_WINDOW (toplevel)) {
+		parent = GTK_WINDOW(toplevel);
+		gtk_window_set_transient_for (GTK_WINDOW(window), parent);
+		gtk_window_set_position (GTK_WINDOW(window), GTK_WIN_POS_CENTER_ON_PARENT);
+	}
+	gtk_widget_show(window);
+	REMMINA_DEBUG ("After popover");
+}
+
 gboolean
 remmina_ssh_plugin_popup_menu(GtkWidget *widget, GdkEvent *event, GtkWidget *menu)
 {
@@ -530,12 +802,13 @@ void remmina_plugin_ssh_popup_ui(RemminaProtocolWidget *gp)
 	/* Context menu for slection and clipboard */
 	GtkWidget *menu = gtk_menu_new();
 
-	GtkWidget *select_all = gtk_menu_item_new_with_label(_("Select All (Host + A)"));
-	GtkWidget *copy = gtk_menu_item_new_with_label(_("Copy (host + C)"));
-	GtkWidget *paste = gtk_menu_item_new_with_label(_("Paste (host + V)"));
+	GtkWidget *select_all = gtk_menu_item_new_with_label(_("Select All (host+A)"));
+	GtkWidget *copy = gtk_menu_item_new_with_label(_("Copy (host+C)"));
+	GtkWidget *paste = gtk_menu_item_new_with_label(_("Paste (host+V)"));
 	GtkWidget *save = gtk_menu_item_new_with_label(_("Save session to file"));
-	GtkWidget *font_incr = gtk_menu_item_new_with_label(_("Increase font size (Host + Page_Up)"));
-	GtkWidget *font_decr = gtk_menu_item_new_with_label(_("Decrease font size (Host + Page_Down)"));
+	GtkWidget *font_incr = gtk_menu_item_new_with_label(_("Increase font size (host+Page Up)"));
+	GtkWidget *font_decr = gtk_menu_item_new_with_label(_("Decrease font size (host+Page Down)"));
+	GtkWidget *find_text = gtk_menu_item_new_with_label(_("Find text (host+G)"));
 
 	gtk_menu_shell_append(GTK_MENU_SHELL(menu), select_all);
 	gtk_menu_shell_append(GTK_MENU_SHELL(menu), copy);
@@ -543,6 +816,7 @@ void remmina_plugin_ssh_popup_ui(RemminaProtocolWidget *gp)
 	gtk_menu_shell_append(GTK_MENU_SHELL(menu), save);
 	gtk_menu_shell_append(GTK_MENU_SHELL(menu), font_incr);
 	gtk_menu_shell_append(GTK_MENU_SHELL(menu), font_decr);
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), find_text);
 
 	g_signal_connect(G_OBJECT(gpdata->vte), "button_press_event",
 			 G_CALLBACK(remmina_ssh_plugin_popup_menu), menu);
@@ -559,6 +833,8 @@ void remmina_plugin_ssh_popup_ui(RemminaProtocolWidget *gp)
 			 G_CALLBACK(remmina_plugin_ssh_vte_increase_font), gpdata->vte);
 	g_signal_connect(G_OBJECT(font_decr), "activate",
 			 G_CALLBACK(remmina_plugin_ssh_vte_decrease_font), gpdata->vte);
+	g_signal_connect(G_OBJECT(find_text), "activate",
+			 G_CALLBACK(remmina_plugin_pop_search), gp);
 
 	gtk_widget_show_all(menu);
 }
@@ -603,7 +879,7 @@ remmina_plugin_ssh_init(RemminaProtocolWidget *gp)
 	g_signal_connect(G_OBJECT(hbox), "focus-in-event", G_CALLBACK(remmina_plugin_ssh_on_focus_in), gp);
 
 	vte = vte_terminal_new();
-	gtk_widget_show(vte);
+	//gtk_widget_show(vte);
 	vte_terminal_set_size(VTE_TERMINAL(vte), 80, 25);
 	vte_terminal_set_scroll_on_keystroke(VTE_TERMINAL(vte), TRUE);
 #if !VTE_CHECK_VERSION(0, 38, 0)
@@ -615,11 +891,15 @@ remmina_plugin_ssh_init(RemminaProtocolWidget *gp)
 #if VTE_CHECK_VERSION(0, 38, 0)
 	GdkRGBA cp[PALETTE_SIZE];
 	GdkRGBA cursor_color;
+	GdkRGBA cursor_foreground;
+	GdkRGBA highlight;
+	GdkRGBA highlight_foreground;
+	GdkRGBA colorBD;
 
 	/*
 	 * custom colors reside inside of the 'theme' subdir of the remmina config folder (.config/remmina/theme)
 	 * with the file extension '.colors'. The name of the colorfile came from the menu (see below)
-	 * sideeffect: it is possible to overwrite the standard colors with a dedicated colorfile like
+	 * sideeffect: It is possible to overwrite the standard colours with a dedicated colourfile like
 	 * '0.colors' for GRUVBOX, '1.colors' for TANGO and so on
 	 */
 	const gchar *color_name = remmina_plugin_service->file_get_string(remminafile, "ssh_color_scheme");
@@ -667,6 +947,10 @@ remmina_plugin_ssh_init(RemminaProtocolWidget *gp)
 		gdk_rgba_parse(&foreground_color, color_pref.foreground);
 		gdk_rgba_parse(&background_color, color_pref.background);
 		gdk_rgba_parse(&cursor_color, color_pref.cursor);
+		gdk_rgba_parse(&cursor_foreground, color_pref.cursor_foreground);
+		gdk_rgba_parse(&highlight, color_pref.highlight);
+		gdk_rgba_parse(&highlight_foreground, color_pref.highlight_foreground);
+		gdk_rgba_parse(&colorBD, color_pref.colorBD);
 
 		gdk_rgba_parse(&cp[0], color_pref.color0);
 		gdk_rgba_parse(&cp[1], color_pref.color1);
@@ -700,42 +984,70 @@ remmina_plugin_ssh_init(RemminaProtocolWidget *gp)
 			gdk_rgba_parse(&foreground_color, "#ffffff");
 			gdk_rgba_parse(&background_color, "#000000");
 			gdk_rgba_parse(&cursor_color, "#ffffff");
+			gdk_rgba_parse(&cursor_foreground, "#00000");
+			gdk_rgba_parse(&highlight, "#ffffff");
+			gdk_rgba_parse(&highlight_foreground, "#00000");
+			gdk_rgba_parse(&colorBD, "#ffffff");
 			remminavte.palette = linux_palette;
 			break;
 		case TANGO:
-			gdk_rgba_parse(&foreground_color, "#eeeeec");
-			gdk_rgba_parse(&background_color, "#2e3436");
-			gdk_rgba_parse(&cursor_color, "#8ae234");
+			gdk_rgba_parse(&foreground_color, "#ffffff");
+			gdk_rgba_parse(&background_color, "#000000");
+			gdk_rgba_parse(&cursor_color, "#000000");
+			gdk_rgba_parse(&cursor_foreground, "#ffffff");
+			gdk_rgba_parse(&highlight, "#ffffff");
+			gdk_rgba_parse(&highlight_foreground, "#00000");
+			gdk_rgba_parse(&colorBD, "#000000");
 			remminavte.palette = tango_palette;
 			break;
 		case GRUVBOX:
-			gdk_rgba_parse(&foreground_color, "#ebdbb2");
-			gdk_rgba_parse(&background_color, "#282828");
-			gdk_rgba_parse(&cursor_color, "#d3869b");
+			gdk_rgba_parse(&foreground_color, "#e6d4a3");
+			gdk_rgba_parse(&background_color, "#1e1e1e");
+			gdk_rgba_parse(&cursor_color, "#e6d4a3");
+			gdk_rgba_parse(&cursor_foreground, "#e6d4a3");
+			gdk_rgba_parse(&highlight, "#e6d4a3");
+			gdk_rgba_parse(&highlight_foreground, "#1e1e1e");
+			gdk_rgba_parse(&colorBD, "#ffffff");
 			remminavte.palette = gruvbox_palette;
 			break;
 		case SOLARIZED_DARK:
 			gdk_rgba_parse(&foreground_color, "#839496");
 			gdk_rgba_parse(&background_color, "#002b36");
 			gdk_rgba_parse(&cursor_color, "#93a1a1");
+			gdk_rgba_parse(&cursor_foreground, "#839496");
+			gdk_rgba_parse(&highlight, "#839496");
+			gdk_rgba_parse(&highlight_foreground, "#002b36");
+			gdk_rgba_parse(&colorBD, "#819090");
 			remminavte.palette = solarized_dark_palette;
 			break;
 		case SOLARIZED_LIGHT:
 			gdk_rgba_parse(&foreground_color, "#657b83");
 			gdk_rgba_parse(&background_color, "#fdf6e3");
 			gdk_rgba_parse(&cursor_color, "#586e75");
+			gdk_rgba_parse(&cursor_foreground, "#657b83");
+			gdk_rgba_parse(&highlight, "#657b83");
+			gdk_rgba_parse(&highlight_foreground, "#fdf6e3");
+			gdk_rgba_parse(&colorBD, "#475b62");
 			remminavte.palette = solarized_light_palette;
 			break;
 		case XTERM:
 			gdk_rgba_parse(&foreground_color, "#000000");
 			gdk_rgba_parse(&background_color, "#ffffff");
 			gdk_rgba_parse(&cursor_color, "#000000");
+			gdk_rgba_parse(&cursor_foreground, "#ffffff");
+			gdk_rgba_parse(&highlight, "#000000");
+			gdk_rgba_parse(&highlight_foreground, "#ffffff");
+			gdk_rgba_parse(&colorBD, "#000000");
 			remminavte.palette = xterm_palette;
 			break;
 		case CUSTOM:
 			gdk_rgba_parse(&foreground_color, remmina_pref.color_pref.foreground);
 			gdk_rgba_parse(&background_color, remmina_pref.color_pref.background);
 			gdk_rgba_parse(&cursor_color, remmina_pref.color_pref.cursor);
+			gdk_rgba_parse(&cursor_foreground, remmina_pref.color_pref.cursor_foreground);
+			gdk_rgba_parse(&highlight, remmina_pref.color_pref.highlight);
+			gdk_rgba_parse(&highlight_foreground, remmina_pref.color_pref.highlight_foreground);
+			gdk_rgba_parse(&colorBD, remmina_pref.color_pref.colorBD);
 
 			gdk_rgba_parse(&cp[0], remmina_pref.color_pref.color0);
 			gdk_rgba_parse(&cp[1], remmina_pref.color_pref.color1);
@@ -773,6 +1085,11 @@ remmina_plugin_ssh_init(RemminaProtocolWidget *gp)
 	vte_terminal_set_color_foreground(VTE_TERMINAL(vte), &foreground_color);
 	vte_terminal_set_color_background(VTE_TERMINAL(vte), &background_color);
 	vte_terminal_set_color_cursor(VTE_TERMINAL(vte), &cursor_color);
+	vte_terminal_set_color_cursor_foreground(VTE_TERMINAL(vte), &cursor_foreground);
+	vte_terminal_set_color_highlight(VTE_TERMINAL(vte), &highlight);
+	vte_terminal_set_color_highlight_foreground(VTE_TERMINAL(vte), &highlight_foreground);
+	vte_terminal_set_color_bold(VTE_TERMINAL(vte), &colorBD);
+
 #else
 	/* VTE <= 2.90 doesn’t support GdkRGBA so we must convert GdkRGBA to GdkColor */
 	foreground_gdkcolor.red = (guint16)(foreground_color.red * 0xFFFF);
@@ -826,6 +1143,7 @@ remmina_plugin_ssh_init(RemminaProtocolWidget *gp)
 	gpdata->vte_session_file = g_file_new_for_path(fp);
 
 	remmina_plugin_ssh_popup_ui(gp);
+	gtk_widget_show_all(hbox);
 }
 
 /**
@@ -947,11 +1265,23 @@ remmina_plugin_ssh_call_feature(RemminaProtocolWidget *gp, const RemminaProtocol
 	case REMMINA_PLUGIN_SSH_FEATURE_TOOL_DECREASE_FONT:
 		vte_terminal_set_font_scale(VTE_TERMINAL(gpdata->vte),
 				vte_terminal_get_font_scale(VTE_TERMINAL(gpdata->vte))-SCALE_FACTOR);
+	case REMMINA_PLUGIN_SSH_FEATURE_TOOL_SEARCH:
+			remmina_plugin_pop_search(NULL, gp);
 		return;
 	}
 }
 
-/** Array of key/value pairs for SSH auth type*/
+/** Array of key/value pairs for SSH auth type
+ * libssh methods:
+ *
+ * #define SSH_AUTH_METHOD_UNKNOWN     0x0000u
+ * #define SSH_AUTH_METHOD_NONE        0x0001u
+ * #define SSH_AUTH_METHOD_PASSWORD    0x0002u
+ * #define SSH_AUTH_METHOD_PUBLICKEY   0x0004u
+ * #define SSH_AUTH_METHOD_HOSTBASED   0x0008u
+ * #define SSH_AUTH_METHOD_INTERACTIVE 0x0010u
+ * #define SSH_AUTH_METHOD_GSSAPI_MIC  0x0020u
+ */
 static gpointer ssh_auth[] =
 {
 	"0", N_("Password"),
@@ -993,22 +1323,22 @@ static gpointer ssh_charset_list[] =
 	"", "EUC-JP",
 	"", "EUC-KR",
 	"", "GBK",
-	"", "ISO-8859-1",
-	"", "ISO-8859-2",
-	"", "ISO-8859-3",
-	"", "ISO-8859-4",
-	"", "ISO-8859-5",
-	"", "ISO-8859-6",
-	"", "ISO-8859-7",
-	"", "ISO-8859-8",
-	"", "ISO-8859-9",
-	"", "ISO-8859-10",
-	"", "ISO-8859-11",
-	"", "ISO-8859-12",
-	"", "ISO-8859-13",
-	"", "ISO-8859-14",
-	"", "ISO-8859-15",
-	"", "ISO-8859-16",
+	"", "ISO 8859-1",
+	"", "ISO 8859-2",
+	"", "ISO 8859-3",
+	"", "ISO 8859-4",
+	"", "ISO 8859-5",
+	"", "ISO 8859-6",
+	"", "ISO 8859-7",
+	"", "ISO 8859-8",
+	"", "ISO 8859-9",
+	"", "ISO 8859-10",
+	"", "ISO 8859-11",
+	"", "ISO 8859-12",
+	"", "ISO 8859-13",
+	"", "ISO 8859-14",
+	"", "ISO 8859-15",
+	"", "ISO 8859-16",
 	"", "KOI8-R",
 	"", "SJIS",
 	"", "UTF-8",
@@ -1038,6 +1368,7 @@ static RemminaProtocolFeature remmina_plugin_ssh_features[] =
 	{ REMMINA_PROTOCOL_FEATURE_TYPE_TOOL, REMMINA_PLUGIN_SSH_FEATURE_TOOL_SELECT_ALL, N_("Select all"), N_("_Select all"), NULL },
 	{ REMMINA_PROTOCOL_FEATURE_TYPE_TOOL, REMMINA_PLUGIN_SSH_FEATURE_TOOL_INCREASE_FONT, N_("Increase font size"), N_("_Increase font size"), NULL },
 	{ REMMINA_PROTOCOL_FEATURE_TYPE_TOOL, REMMINA_PLUGIN_SSH_FEATURE_TOOL_DECREASE_FONT, N_("Decrease font size"), N_("_Decrease font size"), NULL },
+	{ REMMINA_PROTOCOL_FEATURE_TYPE_TOOL, REMMINA_PLUGIN_SSH_FEATURE_TOOL_SEARCH, N_("Find text"), N_("_Find text"), NULL },
 	{ REMMINA_PROTOCOL_FEATURE_TYPE_END,  0,					  NULL,		    NULL,	       NULL }
 };
 
@@ -1055,12 +1386,13 @@ static RemminaProtocolFeature remmina_plugin_ssh_features[] =
 static const RemminaProtocolSetting remmina_ssh_basic_settings[] =
 {
 	{ REMMINA_PROTOCOL_SETTING_TYPE_SERVER,	  "server",	    NULL,				  FALSE, "_ssh._tcp", NULL },
+	{ REMMINA_PROTOCOL_SETTING_TYPE_SELECT,	  "ssh_auth",	    N_("Authentication type"),		  FALSE, ssh_auth,    NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_TEXT,	  "username",	    N_("Username"),			  FALSE, NULL,	      NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_PASSWORD, "password",	    N_("User password"),		  FALSE, NULL,	      NULL },
-	{ REMMINA_PROTOCOL_SETTING_TYPE_SELECT,	  "ssh_auth",	    N_("Authentication type"),		  FALSE, ssh_auth,    NULL },
-	{ REMMINA_PROTOCOL_SETTING_TYPE_FILE,	  "ssh_privatekey", N_("Identity file"),		  FALSE, NULL,	      NULL },
+	{ REMMINA_PROTOCOL_SETTING_TYPE_FILE,	  "ssh_privatekey", N_("SSH identity file"),		  FALSE, NULL,	      NULL },
+	{ REMMINA_PROTOCOL_SETTING_TYPE_FILE,	  "ssh_certfile",   N_("SSH certificate file"),		  FALSE, NULL,	      NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_PASSWORD, "ssh_passphrase", N_("Password to unlock private key"), FALSE, NULL,	      NULL },
-	{ REMMINA_PROTOCOL_SETTING_TYPE_TEXT,	  "exec",	    N_("Startup program"),		  FALSE, NULL,	      NULL },
+	{ REMMINA_PROTOCOL_SETTING_TYPE_TEXT,	  "exec",	    N_("Start-up program"),		  FALSE, NULL,	      NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_END,	  NULL,		    NULL,				  FALSE, NULL,	      NULL }
 };
 
@@ -1072,7 +1404,7 @@ static gchar log_tips[] =
         "  • %U is substituted with the SSH username\n"
         "  • %p is substituted with Remmina profile name\n"
         "  • %g is substituted with Remmina profile group name\n"
-        "  • %d is substituted with local datetime in iso8601 format\n");
+        "  • %d is substituted with local date and time in ISO 8601 format\n");
 
 /**
  * Array of RemminaProtocolSetting for advanced settings.
@@ -1087,7 +1419,7 @@ static gchar log_tips[] =
  */
 static const RemminaProtocolSetting remmina_ssh_advanced_settings[] =
 {
-	{ REMMINA_PROTOCOL_SETTING_TYPE_SELECT, "ssh_color_scheme",	  N_("Terminal color scheme"),		      FALSE, ssh_terminal_palette, NULL },
+	{ REMMINA_PROTOCOL_SETTING_TYPE_SELECT, "ssh_color_scheme",	  N_("Terminal colour scheme"),		      FALSE, ssh_terminal_palette, NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_SELECT, "ssh_charset",		  N_("Character set"),			      FALSE, ssh_charset_list,	   NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_TEXT,	"ssh_proxycommand",	  N_("SSH Proxy Command"),		      FALSE, NULL,		   NULL },
 	{ REMMINA_PROTOCOL_SETTING_TYPE_TEXT,	"ssh_kex_algorithms",	  N_("KEX (Key Exchange) algorithms"),	      FALSE, NULL,		   NULL },
@@ -1245,6 +1577,7 @@ remmina_ssh_plugin_register(void)
 	remmina_plugin_ssh_features[2].opt3 = GUINT_TO_POINTER(remmina_pref.vte_shortcutkey_select_all);
 	remmina_plugin_ssh_features[3].opt3 = GUINT_TO_POINTER(remmina_pref.vte_shortcutkey_increase_font);
 	remmina_plugin_ssh_features[4].opt3 = GUINT_TO_POINTER(remmina_pref.vte_shortcutkey_decrease_font);
+	remmina_plugin_ssh_features[5].opt3 = GUINT_TO_POINTER(remmina_pref.vte_shortcutkey_search_text);
 
 	remmina_plugin_service = &remmina_plugin_manager_service;
 
